@@ -1,6 +1,6 @@
-;;; ghub-graphql.el --- Access Github API using GrapthQL  -*- lexical-binding:t -*-
+;;; ghub-graphql.el --- Access Github API using GraphQL  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2016-2024 Jonas Bernoulli
+;; Copyright (C) 2016-2025 Jonas Bernoulli
 
 ;; Author: Jonas Bernoulli <emacs.ghub@jonas.bernoulli.dev>
 ;; Homepage: https://github.com/magit/ghub
@@ -21,30 +21,37 @@
 ;; You should have received a copy of the GNU General Public License
 ;; along with this file.  If not, see <https://www.gnu.org/licenses/>.
 
+;;; Commentary:
+
+;; This library implements GraphQL queries for Github.
+
 ;;; Code:
 
 (require 'ghub)
 (require 'gsexp)
 (require 'treepy)
 
-;; Needed for Emacs < 27.
-(eval-when-compile (require 'json))
-(declare-function json-read-from-string "json" (string))
-(declare-function json-encode "json" (object))
-
-(eval-when-compile (require 'pp)) ; Needed for Emacs < 29.
 (eval-when-compile (require 'subr-x))
 
-;;; Api
-
 (define-error 'ghub-graphql-error "GraphQL Error" 'ghub-error)
 
+;;; Settings
+
+(defvar ghub-graphql-message-progress nil
+  "Whether to show \"Fetching page N...\" in echo area during requests.
+By default this information is only shown in the mode-line of the buffer
+from which the request was initiated, and if you kill that buffer, then
+nowhere.  That may make it desirable to display the same message in the
+echo area as well.")
+
 (defvar ghub-graphql-items-per-request 50
   "Number of GraphQL items to query for entities that return a collection.
 
 Adjust this value if you're hitting query timeouts against larger
 repositories.")
 
+;;; Mutations
+
 (cl-defun ghub-graphql (graphql
                         &optional variables
                         &key username auth host forge
@@ -70,6 +77,23 @@ behave as for `ghub-request' (which see)."
                 :callback callback :errorback errorback
                 :extra extra :value value))
 
+(cl-defun ghub--graphql (graphql
+                         &optional variables
+                         &key username auth host forge
+                         headers
+                         callback errorback)
+  "An experimental and unfinished replacement for `ghub-graphql'."
+  (let ((ghub-graphql-message-progress nil))
+    (ghub--graphql-vacuum graphql variables callback nil
+                          :username  username
+                          :auth      auth
+                          :host      host
+                          :forge     forge
+                          :headers   headers
+                          :errorback errorback)))
+
+;;; Queries
+
 (cl-defun ghub-graphql-rate-limit (&key username auth host)
   "Return rate limit information."
   (let-alist (ghub-graphql
@@ -77,19 +101,6 @@ behave as for `ghub-request' (which see)."
               nil :username username :auth auth :host host)
     .data.rateLimit))
 
-(cl-defun ghub--repository-id (owner name &key username auth host)
-  "Return the id of the repository specified by OWNER, NAME and HOST."
-  (let-alist (ghub-graphql
-              '(query (repository [(owner $owner String!)
-                                   (name  $name  String!)]
-                                  id))
-              `((owner . ,owner)
-                (name  . ,name))
-              :username username :auth auth :host host)
-    .data.repository.id))
-
-;;; Api (drafts)
-
 (defconst ghub-fetch-repository-sparse
   '(query
     (repository
@@ -109,6 +120,7 @@ behave as for `ghub-request' (which see)."
      isLocked
      isMirror
      isPrivate
+     hasDiscussionsEnabled
      hasIssuesEnabled
      hasWikiEnabled
      (licenseInfo name)
@@ -123,6 +135,46 @@ behave as for `ghub-request' (which see)."
                       id
                       login
                       name)
+     (discussionCategories [(:edges t)]
+                            id
+                            name
+                            emoji
+                            isAnswerable
+                            description)
+     (discussions    [(:edges t)
+                      (:singular discussion number)
+                      (orderBy ((field UPDATED_AT) (direction DESC)))]
+                     id
+                     databaseId
+                     number
+                     url
+                     stateReason
+                     ;; Discussions lack isReadByViewer.
+                     (answer id)
+                     (author login)
+                     title
+                     createdAt
+                     updatedAt
+                     closedAt
+                     locked
+                     (category id)
+                     body
+                     (comments  [(:edges t)]
+                                id
+                                databaseId
+                                (author login)
+                                createdAt
+                                updatedAt
+                                body
+                                (replies [(:edges 20)]
+                                         id
+                                         databaseId
+                                         (author login)
+                                         createdAt
+                                         updatedAt
+                                         body))
+                     (labels    [(:edges t)]
+                                id))
      (issues         [(:edges t)
                       (:singular issue number)
                       (orderBy ((field UPDATED_AT) (direction DESC)))]
@@ -256,6 +308,27 @@ data as the only argument."
                         :paginate paginate
                         :errorback errorback))
 
+(cl-defun ghub-fetch-discussion ( owner name number callback
+                                  &optional until
+                                  &key username auth host forge
+                                  headers errorback)
+  "Asynchronously fetch forge data about the specified discussion.
+Once all data has been collected, CALLBACK is called with the
+data as the only argument."
+  (ghub--graphql-vacuum (ghub--graphql-prepare-query
+                         ghub-fetch-repository
+                         `(repository discussions (discussion . ,number)))
+                        `((owner . ,owner)
+                          (name  . ,name))
+                        callback until
+                        :narrow   '(repository discussion)
+                        :username username
+                        :auth     auth
+                        :host     host
+                        :forge    forge
+                        :headers  headers
+                        :errorback errorback))
+
 (cl-defun ghub-fetch-issue ( owner name number callback
                              &optional until
                              &key username auth host forge
@@ -324,9 +397,6 @@ data as the only argument."
 
 ;;; Internal
 
-(defvar ghub--graphql-debug nil
-  "Whether `ghub--graphql-retrieve' updates the \" *gsexp-encode*\" buffer.")
-
 (cl-defstruct (ghub--graphql-req
                (:include ghub--req)
                (:constructor ghub--make-graphql-req)
@@ -390,6 +460,9 @@ See Info node `(ghub)GraphQL Support'."
 
 (cl-defun ghub--graphql-retrieve (req &optional lineage cursor)
   (let ((p (cl-incf (ghub--graphql-req-pages req))))
+    (when ghub-graphql-message-progress
+      (let ((message-log-max nil))
+        (message "Fetching page %s..." p)))
     (when (> p 1)
       (ghub--graphql-set-mode-line req "Fetching page %s" p)))
   (setf (ghub--graphql-req-query-str req)
@@ -397,10 +470,13 @@ See Info node `(ghub)GraphQL Support'."
          (ghub--graphql-prepare-query
           (ghub--graphql-req-query req)
           lineage cursor)))
-  (when ghub--graphql-debug
+  (when ghub-debug
     (with-current-buffer (get-buffer-create " *gsexp-encode*")
       (erase-buffer)
-      (insert (ghub--graphql-req-query-str req))))
+      (insert (ghub--graphql-req-query-str req) "\n\n")
+      (let ((pos (point)))
+        (insert (ghub--encode-payload (ghub--graphql-req-variables req)) "\n")
+        (ignore-errors (json-pretty-print pos (point))))))
   (ghub--retrieve
    (let ((json-false nil))
      (ghub--encode-payload
@@ -464,11 +540,15 @@ See Info node `(ghub)GraphQL Support'."
                  (err     (plist-get status :error))
                  (errors  (cdr (assq 'errors payload)))
                  (errors  (and errors (cons 'ghub-graphql-error errors))))
-            (if (or err errors)
-                (ghub--graphql-handle-failure
-                 req (or err errors) headers status)
-              (ghub--graphql-walk-response req (assq 'data payload)))))
-      (when (buffer-live-p buf)
+            (cond ((or err errors)
+                   (when (and (not err) ghub-debug)
+                     (ignore-errors (json-pretty-print (point) (point-max)))
+                     (pop-to-buffer buf))
+                   (ghub--graphql-handle-failure
+                    req (or err errors) headers status))
+                  ((ghub--graphql-walk-response req (assq 'data payload))))))
+      (when (and (buffer-live-p buf)
+                 (not (buffer-local-value 'ghub-debug buf)))
         (kill-buffer buf)))))
 
 (defun ghub--graphql-handle-failure (req errors headers status)
@@ -594,7 +674,7 @@ See Info node `(ghub)GraphQL Support'."
                 child))))))
 
 (defun ghub--alist-zip (root)
-  (let ((branchp (lambda (elt) (and (listp elt) (listp (cdr elt)))))
+  (let ((branchp (##and (listp %) (listp (cdr %))))
         (make-node (lambda (_ children) children)))
     (treepy-zipper branchp #'identity make-node root)))
 
@@ -607,9 +687,19 @@ See Info node `(ghub)GraphQL Support'."
         (force-mode-line-update t)))))
 
 (defun ghub--graphql-pp-response (data)
-  (require 'pp) ; needed for Emacs < 29.
   (pp-display-expression data "*Pp Eval Output*"))
 
+(cl-defun ghub--repository-id (owner name &key username auth host)
+  "Return the id of the repository specified by OWNER, NAME and HOST."
+  (let-alist (ghub-graphql
+              '(query (repository [(owner $owner String!)
+                                   (name  $name  String!)]
+                                  id))
+              `((owner . ,owner)
+                (name  . ,name))
+              :username username :auth auth :host host)
+    .data.repository.id))
+
 ;;; _
 (provide 'ghub-graphql)
 ;;; ghub-graphql.el ends here
